/*
 * libtota
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 *******************************************************************************
 * \file	SS_FileSystemUpdate.h
 *
 * \brief	UPI FS Update API
 *******************************************************************************
 */
#ifndef _SS_FILESYSTEM_UPDATE_H_
#define _SS_FILESYSTEM_UPDATE_H_

#include "SS_Engine_Update.h"
#include "SS_Engine_Errors.h"

#ifdef __cplusplus
extern "C" {
#endif                          /* __cplusplus */

/*!
 * File access modes
 */
	typedef enum tag_RW_TYPE {
		ONLY_R,                 //!< Read-only
		ONLY_W,                 //!< Write-only
		BOTH_RW                 //!< Read-write
	} E_RW_TYPE;

/*!
 *******************************************************************************
 * Copy file.<p>
 *
 * Must create the path to the new file as required. Must overwrite any contents
 * in the old file, if any. Must check if the source file is a symbolic link.
 * If it is, instead create a new symbolic link using \ref SS_Link.
 *
 * \param	strFromPath	Path to old file
 * \param	strToPath	Path to new file
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_CopyFile(const char *strFromPath, const char *strToPath);

/*!
 *******************************************************************************
 * Move (rename) file.<p>
 *
 * \param	strFromPath	Path to old file location
 * \param	strToPath	Path to new file location
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_MoveFile(const char *strFromPath, const char *strToPath);

/*!
 *******************************************************************************
 * Delete file.<p>
 *
 * Must return success if the file is deleted or didn't exist.
 *
 * \param	strPath		Path to file
 *
 * \return	S_SS_SUCCESS on success, E_SS_DELETEFILE if the file cannot be
 *			deleted, or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_DeleteFile(const char *strPath);

/*!
 *******************************************************************************
 * Delete folder.<p>
 *
 * Must return success if the folder is now deleted or didn't exist.
 *
 * \param	strPath		Path to folder
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_DeleteFolder(const char *strPath);

/*!
 *******************************************************************************
 * Create folder.<p>
 *
 * Must return success if the folder is created or already existsed. It is
 * recommended that the new folder's attributes are a copy of its parent's
 * attributes.
 *
 * \param	strPath		Path to folder
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_CreateFolder(const char *strPath);

/*!
 *******************************************************************************
 * Open file.<p>
 *
 * Must create the the file (and the path to the file) if it doesn't exist. Must
 * open in binary mode.
 *
 * \param	strPath		Path to file
 * \param	wFlag		Read/write mode, an \ref E_RW_TYPE value
 * \param	pwHandle	(out) File handle
 *
 * \return	S_SS_SUCCESS on success, E_SS_OPENFILE_ONLYR if attempting to open a
 *			non-existant file in R/O mode, E_SS_OPENFILE_WRITE if there is an
 *			error opening a file for writing, or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_OpenFile(const char *strPath, E_RW_TYPE wFlag, long *pwHandle);

/*!
 *******************************************************************************
 * Set file size.
 *
 * \param	wHandle		File handle
 * \param	dwSize		New file size
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_ResizeFile(long wHandle, SS_UINT32 dwSize);

/*!
 *******************************************************************************
 * Close file.
 *
 * \param	wHandle		File handle
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_CloseFile(long wHandle);

/*!
 *******************************************************************************
 * Write data to a specified position within a file.<p>
 *
 * Must return success if the block is written or at least resides in
 * non-volatile memory. Use \ref SS_SyncFile to commit the file to storage.
 *
 * \param	wHandle		File handle
 * \param	dwPosition	Position within the file to which to write
 * \param	pbBuffer	Data to write
 * \param	dwSize		Size of \a pbBuffer
 *
 * \return	S_SS_SUCCESS on success, or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_WriteFile(long wHandle, SS_UINT32 dwPosition, unsigned char *pbBuffer, SS_UINT32 dwSize);

/*!
 *******************************************************************************
 * Commit file to storage.<p>
 *
 * Generally called after \ref SS_WriteFile.
 *
 * \param	wHandle		File handle
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_SyncFile(long wHandle);

/*!
 *******************************************************************************
 * Read data from a specified position within a file.
 * If fewer bytes than requested are available in the specified position, this
 * function should read up to the end of file and return S_SS_SUCCESS.
 *
 * \param	wHandle		File handle
 * \param	dwPosition	Position within the file from which to read
 * \param	pbBuffer	Buffer to contain data
 * \param	dwSize		Size of data to read
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */

	long SS_ReadFile(long wHandle, SS_UINT32 dwPosition, unsigned char *pbBuffer, SS_UINT32 dwSize);

/*!
 *******************************************************************************
 * Get file size.
 *
 * \param	wHandle		File handle
 *
 * \return	File size, -1 if file not found, or &lt; -1 on error
 *******************************************************************************
 */
	long SS_GetFileSize(long wHandle);

/*!
 *******************************************************************************
 * Get free space of a mounted file system.
 *
 * \param	path				    Name of any directory within the mounted
 *									file system
 * \param	available_flash_size	(out) Available space
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_GetAvailableFreeSpace(const char *path, SS_UINT32 * available_flash_size);

/*!
 *******************************************************************************
 * Remove symbolic link.<p>
 *
 * Must return an error if the file does not exist or is not a symbolic link.<p>
 *
 * If your platform does not support symbolic links, you do not need to
 * implement this.
 *
 * \param	pLinkName	Link
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_Unlink(char *pLinkName);

/*!
 *******************************************************************************
 * Create symbolic link.<p>
 *
 * Must create the path to the link as required. If a file already exists at the
 * named location, must return success if the file is a symbolic link or an
 * error if the file is a regular file. The non-existance of the target of the
 * link must NOT cause an error.<p>
 *
 * If your platform does not support symbolic links, you do not need to
 * implement this.
 *
 * \param	pbUserData			Optional opaque data-structure to pass to IPL
 *								functions
 * \param	pLinkName			Path to the link file to create
 * \param	pReferenceFileName	Path to which to point the link
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_Link(void *pbUserData, char *pLinkName, char *pReferenceFileName);

/*!
 *******************************************************************************
 * Set file attributes.<p>
 *
 * The file attributes token (\a ui8pAttribs) is defined at generation time.
 * If attributes are not defined explicitly, they are given the following,
 * OS-dependent values:
 * \li	Windows: _foo_ro_ for R/O files, _foo_rw_ for R/W files
 * \li	Linux: _foo_oooooo:xxxx:yyyy indicating the file mode, uid, and gid
 *		(uid and gid use capitalized hex digits as required)
 *
 * \param	pbUserData		Optional opaque data-structure to pass to IPL
 *							functions
 * \param	ui16pFilePath	File path
 * \param	ui32AttribSize	Size of \a ui8pAttribs
 * \param	ui8pAttribs		Attributes to set
 *
 * \return	S_SS_SUCCESS on success or &lt; 0 on error
 *******************************************************************************
 */

	long SS_SetFileAttributes(const char *ui16pFilePath,
			const SS_UINT32 ui32AttribSize, const unsigned char *ui8pAttribs);

/*!
 *******************************************************************************
 * Print status and debug information.
 *
 * \param	aFormat		A NULL-terminated printf-like string with support for
 *						the following tags:
 *						\li %x:  Hex number
 *						\li %0x: Hex number with leading zeros
 *						\li %u:  Unsigned decimal
 *						\li %s:  NULL-terminated string
 * \param	...			Strings to insert in \a aFormat
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	SS_UINT32 SS_Trace(const char *aFormat, ...);

/**
 *******************************************************************************
 * Allocate a memory block.
 *
 * \param	size	Memory block size, in blocks
 *
 * \return	A pointer to the memory block on success, NULL on failure
 *******************************************************************************
 */
	void *SS_Malloc(SS_UINT32 size);

/**
 *******************************************************************************
 * Free a memory block.
 *
 * \param	pMemBlock	Pointer to the memory block
 *******************************************************************************
 */
	void SS_Free(void *pMemBlock);

/**
 *******************************************************************************
 * Get capability feature flag value
 *
 * \return	The value of feature_support_capability
 *******************************************************************************
 */
	int SS_get_feature_support_capability(void);

/**
 *******************************************************************************
 * Set capability feature flag
 *
 * \param	val	The value to set feature_support_capability
 *******************************************************************************
 */
	void SS_set_feature_support_capability(int val);

#ifdef __cplusplus
}
#endif                          /* __cplusplus */
#endif
