/*
 * libtota
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 *******************************************************************************
 * \file	SS_MultiProcessUpdate.h
 *
 * \brief	UPI MultiProcess API
 *
 * Multi-process updates can work on platforms that support multi-processing.
 * To use the UPI with multi-process support you should implement the following
 * Porting Layer functions:
 *		SS_Malloc
 *		SS_Free
 *		SS_WaitForProcess
 *		SS_RunProcess
 *		SS_GetMaxNumProcess
 *		SS_GetMaxProcRamSize
 *
 * The function SS_RunProcess must run a sub-process (optionally the same
 * executable running the main process) that is also integrated with the UPI.
 * This process must call the UPI API function SS_HandleProcessRequest.
 * <p>
 * The initial call to SS_vRM_Update does not change. vRapid Mobile calls
 * SS_GetMaxNumProcess to see if multi-processing is supported and to see how
 * many processes are supported. If SS_GetMaxNumProcess returns a value greater
 * than zero, the UPI will run parts of the update in sub-processes.
 * <p>
 * The UPI creates a sub-process by calling SS_RunProcess with the arguments
 * for the request. The new process must call SS_HandleProcessRequest to
 * handle the request. The UPI manages processes using SS_WaitForProcess.
 * Each sub-process allocates and frees memories using SS_Malloc and SS_Free.
 * <p>
 * Notes:
 * \li	In the function SS_RunProcess it is recommended to flush output streams
 *		(like stdout or log files) before creating the child process to prevent
 *		buffer duplication in the child process.
 * \li	SS_HandleProcessRequests starts a UPI operation that requires all
 *		standard and common IPL functions in SS_vRM_FileSystemUpdate.h and
 *		SS_vRM_Update.h
 * \li	If your implementation of these functions requires any preparations,
 *		such as opening the update file for SS_GetDelta or opening a log file
 *		for SS_Trace, these preparations must be done before calling
 *		SS_HandleProcessRequests.
 * \li  To distinguish between processes when running in Multi-Process Update mode,
 *		it is recommended that you add the Process ID (i.e. getpid()) to the log
 *		output in SS_Trace. Alternatively, create a log file for each sub-process
 *		named log_file.[pid]
 * \li	If your implementation uses the optional opaque data structure ('user'
 *		in the functions below), you may also have to pass it to sub-processes.
 *		You must implement this functionality.
 *******************************************************************************
 */
#ifndef _SSENGINE_MULTIPROCESS_UPDATE_H_
#define _SSENGINE_MULTIPROCESS_UPDATE_H_

#include "SS_Engine_Update.h"
#include "SS_Engine_Errors.h"

#ifdef __cplusplus
extern "C" {
#endif                          /* __cplusplus */

/**
 *******************************************************************************
 * Allocate a memory block.
 *
 * \param	size	Memory block size, in blocks
 *
 * \return	A pointer to the memory block on success, NULL on failure
 *******************************************************************************
 */
	void *SS_Malloc(SS_UINT32 size);

/**
 *******************************************************************************
 * Free a memory block.
 *
 * \param	pMemBlock	Pointer to the memory block
 *******************************************************************************
 */
	void SS_Free(void *pMemBlock);

/**
 *******************************************************************************
 * Wait for a process to complete.
 *
 * \param	handle				Process handle. If NULL, wait for any process to
 *								complete.
 * \param	process_exit_code	Exit code of the completed process
 *
 * \return	Handle of the completed process or NULL on error
 *******************************************************************************
 */
	void *SS_WaitForProcess(const void *handle, SS_UINT32 * process_exit_code);

/**
 *******************************************************************************
 * Create a new process.
 *
 * The new process must call \ref SS_HandleProcessRequest with the same
 * arguments.
 *
 * \return	Handle to the new process or NULL on error
 *******************************************************************************
 */
	void *SS_RunProcess(void);

/**
 *******************************************************************************
 * Get the maximum number of processes that are allowed to run.
 *
 * \return	The number of allowed processes or 0 if multiple processes are not
 *			allowed
 *******************************************************************************
 */
	SS_UINT32 SS_GetMaxNumProcess(void);
/**
 *******************************************************************************
 * Get the maximum available memory for processes to use.
 *
 * \return	The memory amount available for processes or 0 if there is no available memory
 *******************************************************************************
 */
	unsigned long SS_GetMaxProcRamSize(void);

/**
 *******************************************************************************
 * Initialize a new process and handle the request sent from an external
 * process.
 *
 * Called from the main function of the new process with the parameters passed to
 * \ref SS_RunProcess.
 *
 * \param	user	Optional opaque data-structure passed in during
 *					initialization
 * \param	argc	Number of \a argv arguments
 * \param	argv	Arguments to pass to the new process
 *
 * \return	S_SS_SUCCESS on success or an error code
 *******************************************************************************
 */
	long SS_HandleProcessRequest(void *user, int argc, char *argv[]);

#ifdef __cplusplus
}
#endif                          /* __cplusplus */
#endif
